==> core_extended-111.06.00.tar.gz: core_extended-111.06.00/lib/extended_string.ml <==
open Core.Std;;

(* Natural ordering like found in gnome nautilus, the mac finder etc...
   Refer to Mli for more documentation
*)
let collate s1 s2 =
  let pos1 = ref 0
  and pos2 = ref 0  in

  let next ~ok s pos =
    if (!pos) = String.length s then
      None
    else
      let c = s.[!pos] in
      if ok c then begin
        incr pos;
        Some c
      end else
        None
  in

  let compare_non_numerical () =
    let ok c = not (Char.is_digit c) in
    let rec loop () =
      match next ~ok s1 pos1,next ~ok s2 pos2 with
      | Some _, None -> 1
      | None , Some _ -> -1
      | None , None -> 0
      | Some c1,Some c2 when c1 = c2 -> loop ()
      | Some c1,Some c2 -> Char.compare c1 c2
    in
    loop ()
  in

  let compare_numerical () =
    let rec consume0 s pos  =
      match next ~ok:((=) '0') s pos with
      | Some _ -> consume0 s pos
      | None ->  ()
    in
     (* Our main loop works on string representation of ints where all the
        trailing zeros have been chopped of. Their magnitude is given by the
        length of their representation. If they have the same magnitude the
        lexical order is correct. Bias is used to save that information.
     *)
    let ok = Char.is_digit in
    let bias = ref 0 in
    let rec loop () =
      match next ~ok s1 pos1,next ~ok s2 pos2 with
      | Some _, None -> 1
      | None , Some _ -> -1
      | None , None when !bias <> 0-> !bias
      | None , None ->
          (* Both ints have the same value, The one with the shortest
             representation (i.e. the least trailing zeroes) is
             considered to be the smallest*)
          !pos1 - !pos2
      | Some c1,Some c2 when !bias = 0 ->
          bias := Char.compare c1 c2;
          loop ()
      | Some _ , Some _ -> loop ()
    in
    consume0 s1 pos1;
    consume0 s2 pos2;
    loop ()
  in

  let s1_length = String.length s1 in
  let s2_length = String.length s2 in
  let rec loop () =
    let r  = compare_non_numerical () in
    let r' = compare_numerical () in
    match r,r' with
    | 0,0 when !pos1 = s1_length && !pos2 = s2_length -> 0
    | 0,0 -> loop ()
    | 0,i | i,_ -> i
  in
  loop ()
;;

TEST_MODULE "collate" = struct
  let (<!) s s' = collate s s' < 0

(*
  let (>!) s s' = collate s s' > 0

  let basic_tests = (fun (s,s') ->
    "invertible" @? ((s' <! s) = (s >! s'));
    "total" @? (definitive_clause [s<!s'; s=s'; s>!s']))
*)

  (* repeat 50 basic_tests (pg sg sg);
  repeat 2 basic_tests (dup sg);
  repeat 50 (fun (s,s',s'') ->
    let (s1,s2,s3) =
      match List.sort ~cmp:String.collate [s;s';s''] with
      | [s1;s2;s3] -> s1,s2,s3
      | _ -> assert false
    in
    "transitive" @?
      (((s1 <! s2) || (s2 <! s3)) = (s1 <! s3)))
    (tg sg sg sg); *)

  TEST = "a2b" <! "a10b"
  TEST = "a2b" <! "a02b"
  TEST = "a010b" <! "a20b"

end

(**
   Inverse operation of [String.escaped]
*)
exception Unescape_error of bool*int*string

(* The stdlib's escaped does a lot of fancy wazoo magic to avoid
   using a buffer:
   It works in two passes, the first one calculates the length of the string to
   allocate and the second one does the actual escaping.

   This would be more cumbersome to do here but might be worth the hassle if
   performance ever gets to be an issue *)
let unescaped' ?(strict=true) s =
  let len = String.length s in
  let pos = ref 0 in
  let error ?(fatal=false) message =
    raise (Unescape_error (fatal,!pos,message))
  in
  let consume () =
    let i = !pos in
    if i = len then error "unexpectedly reached end of string";
    let c = s.[i] in
    pos := i + 1;
    c
  in
  let res = Buffer.create len in
  let emit c = Buffer.add_char res c in
  let emit_code code =
    match Char.of_int code with
    | Some c -> emit c
    | None -> error ~fatal:true
        (Printf.sprintf "got invalid escape code %d" code)
  in
  let rec loop () =
    if !pos < len then begin
      let c = consume () in
      if c <> '\\' then
        emit c
      else begin
        let mark = !pos in
        try
          let c = consume () in
          match c with
          | '\\' | '\"' -> emit c
          | 'b' -> emit '\b'
          | 'n' -> emit '\n'
          | 'r' -> emit '\r'
          | 't' -> emit '\t'
          | '\n' ->
              let rec consume_blank () =
                if !pos < len then begin
                  match consume () with
                  | ' ' | '\t' -> consume_blank ()
                  | _ -> decr pos
                end
              in
              consume_blank ()
          | 'x' ->
              let c2hex c =
                if (c >= 'A') && (c <= 'F' ) then
                  (Char.to_int c) + 10 - Char.to_int 'A'
                else if (c >= 'a') && (c <= 'f' ) then
                  (Char.to_int c) + 10 - Char.to_int 'a'
                else if (c >= '0') && (c <= '9') then
                  (Char.to_int c) - Char.to_int '0'
                else
                  error (Printf.sprintf "expected hex digit, got: %c" c);
              in
              let c1 = consume () in
              let c2 = consume () in
              emit_code (16 * c2hex c1 + c2hex c2);
          | c when Char.is_digit c ->
              let char_to_num c =
                match Char.get_digit c with
                | None -> error (Printf.sprintf "expected digit,got: %c" c);
                | Some i -> i
              in
              let i1 = char_to_num c in
              let i2 = char_to_num (consume ()) in
              let i3 = char_to_num (consume ()) in
              emit_code (100 * i1 + 10 * i2 + i3);
          | c -> error (Printf.sprintf "got invalid escape character: %c" c);
        with Unescape_error (false,_,_) when not strict ->
          emit '\\';
          pos := mark
      end;
      loop ()
    end else
      Buffer.contents res;
  in
  loop ();
;;

let unescaped ?strict s =
  try
    unescaped' ?strict s
  with Unescape_error (_,pos,message) ->
    invalid_argf "String.unescaped error at position %d of %s: %s"
      pos s message ()

let unescaped_res ?strict s =
  try
    Core.Result.Ok (unescaped' ?strict s)
  with Unescape_error (_,pos,message) ->
    Core.Result.Error (pos,message)


let squeeze str =
  let len = String.length str in
  let buf = Buffer.create len in
  let rec skip_spaces i =
    if i >= len then
      Buffer.contents buf
    else
      let c = str.[i] in
      if (c = ' ') || (c = '\n') || (c = '\t') || (c = '\r') then
        skip_spaces (i+1)
      else
        begin
          Buffer.add_char buf c;
          copy_chars (i+1)
        end
  and copy_chars i =
    if i >= len then
      Buffer.contents buf
    else
      let c = str.[i] in
      if (c = ' ') || (c = '\n') || (c = '\t') || (c = '\r') then
        begin
          Buffer.add_char buf ' ';
          skip_spaces (i+1)
        end
      else
        begin
          Buffer.add_char buf c;
          copy_chars (i+1)
        end
  in
  copy_chars 0
;;


let pad_right ?(char=' ') s l =
  let src_len = String.length s in
  if src_len >= l then
    s
  else
    let res = String.create l in
    String.blit ~src:s ~dst:res ~src_pos:0 ~dst_pos:0 ~len:src_len;
    String.fill ~pos:src_len ~len:(l-src_len) res char;
    res

let pad_left ?(char=' ') s l =
  let src_len = String.length s in
  if src_len >= l then
    s
  else
    let res = String.create l in
    String.blit ~src:s ~dst:res ~src_pos:0 ~dst_pos:(l-src_len) ~len:src_len;
    String.fill ~pos:0 ~len:(l-src_len) res char;
    res

let line_break ~len s =
  let buf = Buffer.create len in
  let flush_buf () =
    let res = Buffer.contents buf in
    Buffer.reset buf;
    res
  in
  let rec loop acc = function
    | [] ->
        let acc = if Buffer.length buf <> 0 then
          flush_buf ():: acc
        else if acc = [] then
          [""]
        else
          acc
        in
        List.rev acc
    | h::t when Buffer.length buf = 0 ->
        Buffer.add_string buf h;
        loop acc t
    | h::t when (Buffer.length buf + 1 + String.length h) < len ->
        Buffer.add_char buf ' ';
        Buffer.add_string buf h;
        loop acc t
    | l ->
        loop (flush_buf ()::acc) l
  in
  List.concat_map (String.split ~on:'\n' s)
    ~f:(fun s -> loop [] (String.split ~on:' ' s))

(* Finds out where to break a given line; returns the len of the line to break
    and the staring position of the next line.*)
let rec word_wrap__break_one
    ~hard_limit
    ~soft_limit
    ~previous_match
    s
    ~pos
    ~len =
  if pos = String.length s then
    len,pos
  else if previous_match > 0 && len >= soft_limit then
    previous_match,pos-len+previous_match+1
  else if len >= hard_limit then
    len,pos
  else
    match s.[pos] with
      (* Detect \r\n as one newline and not two... *)
    | '\r' when pos < String.length s -1 && s.[pos + 1] = '\n' ->
        len,pos+2
    | '\r' | '\n' -> len,pos+1
    | ' ' | '\t' ->
        word_wrap__break_one s
          ~hard_limit
          ~soft_limit
          ~previous_match:len
          ~pos:(pos+1)
          ~len:(len+1)
    | _ ->
        word_wrap__break_one s
          ~previous_match
          ~hard_limit
          ~soft_limit
          ~pos:(pos+1)
          ~len:(len+1)


(* Returns an pos*length list of all the lines (as substrings of the argument
   passed in) *)
let rec word_wrap__find_substrings ~hard_limit ~soft_limit s acc pos =
  if pos < String.length s then begin
    let len,new_pos = word_wrap__break_one s
      ~hard_limit
      ~soft_limit
      ~previous_match:0
      ~pos
      ~len:0
    in
    word_wrap__find_substrings
      ~hard_limit
      ~soft_limit
      s
      ((pos,len)::acc)
      new_pos
  end else
    acc

let word_wrap
    ?(trailing_nl=false)
    ?(soft_limit=80)
    ?(hard_limit=Int.max_value)
    ?(nl="\n") s =
  let soft_limit = min soft_limit hard_limit in
  let lines = word_wrap__find_substrings ~soft_limit ~hard_limit s [] 0 in
  match lines with
  | [] | [_] ->
      if trailing_nl then s^nl else s
  | ((hpos,hlen)::t) ->
      let nl_len = String.length nl in
      let body_len =
        List.fold_left t
          ~f:(fun acc (_,len) -> acc + nl_len + len)
          ~init:0
      in
      let res_len =
        if trailing_nl then body_len+hlen+nl_len else body_len+hlen
      in
      let res = String.create res_len in
      if trailing_nl then begin
        String.blit
          ~src:nl
          ~dst:res
          ~len:nl_len
          ~src_pos:0
          ~dst_pos:(body_len+hlen);
      end;
      String.blit
        ~src:s
        ~dst:res
        ~len:hlen
        ~src_pos:hpos
        ~dst_pos:body_len;
      let rec blit_loop dst_end_pos = function
        | [] -> ()
        | (src_pos,len)::rest ->
            let dst_pos = dst_end_pos-len-nl_len in
            String.blit
              ~src:s
              ~dst:res
              ~len
              ~src_pos
              ~dst_pos;
            String.blit
              ~src:nl
              ~dst:res
              ~len:nl_len
              ~src_pos:0
              ~dst_pos:(dst_pos + len);
            blit_loop dst_pos rest
      in
      blit_loop body_len t;
      res

(* Knuth-Morris-Pratt string matching. *)
let is_substring  ~substring t =
  let kmp_prefix len ~substring =
    let prefix = Array.create ~len:len 0 in
    let rec f ~k ~q =
      if q > len then prefix
      else (
        let k =
          let rec g k =
            if k <= 0 || ((String.get substring k) = (String.get substring (q - 1))) then k
            else g (prefix.(k - 1))
          in
          g k
        in
        let k =
          if (String.get substring k) = (String.get substring (q - 1))
          then k + 1 else k
        in
        assert (q - 1 >= 0 && q - 1 < len);
        Array.set prefix (q - 1) k;
        f ~k ~q:(q + 1)
      )
    in
    f ~k:0 ~q:2
  in
  let n = String.length t in
  let m = String.length substring in
  let prefix = kmp_prefix m ~substring in
  let rec f ~q ~i =
    if i > n then false
    else
    let q = (
      let q =
        let rec g q =
          if q <= 0 || ((String.get substring q) = (String.get t (i - 1)))
          then q
          else g (prefix.(q - 1))
        in
        g q
      in
      if String.get substring q = String.get t (i - 1) then q + 1
      else q
    )
    in
    if q = m then true
    else f ~q ~i:(i + 1)
  in
  f ~q:0 ~i:1


let edit_distance_matrix ?transpose s1 s2 =
  let transpose = Option.is_some transpose in
  let l1, l2 = String.length s1, String.length s2 in
  let d = Array.make_matrix 0 ~dimx:(l1+1) ~dimy:(l2+1) in
  for x=0 to l1 do d.(x).(0) <- x done;
  for y=0 to l2 do d.(0).(y) <- y done;
  for y=1 to l2 do
    for x=1 to l1 do
      let min_d =
        if s1.[x-1] = s2.[y-1] then d.(x-1).(y-1)
        else List.reduce_exn ~f:min
          [d.(x-1).(y) + 1;
           d.(x).(y-1) + 1;
           d.(x-1).(y-1) + 1]
      in
      let min_d =
        if transpose && x > 1 && y > 1
          && s1.[x-1] = s2.[y-2] && s1.[x-2] = s2.[y-1]
        then min min_d (d.(x-2).(y-2) + 1)
        else min_d
      in
      d.(x).(y) <- min_d
    done;
  done;
  d

let edit_distance ?transpose s1 s2 =
  (edit_distance_matrix ?transpose s1 s2).(String.length s1).(String.length s2)

TEST = edit_distance "" "" = 0
TEST = edit_distance "stringStringString" "stringStringString" = 0
TEST = edit_distance "ocaml" "coaml" = 2
TEST = edit_distance ~transpose:() "ocaml" "coaml" = 1
TEST = edit_distance "sitting" "kitten" = 3
TEST = edit_distance ~transpose:() "sitting" "kitten" = 3
TEST = edit_distance "abcdef" "1234567890" = 10
TEST = edit_distance "foobar" "fubahr" = 3
TEST = edit_distance "hylomorphism" "zylomorphism" = 1

==> core_extended-112.06.00.tar.gz: core_extended-112.06.00/lib/extended_string.ml <==
open Core.Std;;

(* Natural ordering like found in gnome nautilus, the mac finder etc...
   Refer to Mli for more documentation
*)
let collate s1 s2 =
  let pos1 = ref 0
  and pos2 = ref 0  in

  let next ~ok s pos =
    if (!pos) = String.length s then
      None
    else
      let c = s.[!pos] in
      if ok c then begin
        incr pos;
        Some c
      end else
        None
  in

  let compare_non_numerical () =
    let ok c = not (Char.is_digit c) in
    let rec loop () =
      match next ~ok s1 pos1,next ~ok s2 pos2 with
      | Some _, None -> 1
      | None , Some _ -> -1
      | None , None -> 0
      | Some c1,Some c2 when c1 = c2 -> loop ()
      | Some c1,Some c2 -> Char.compare c1 c2
    in
    loop ()
  in

  let compare_numerical () =
    let rec consume0 s pos  =
      match next ~ok:((=) '0') s pos with
      | Some _ -> consume0 s pos
      | None ->  ()
    in
     (* Our main loop works on string representation of ints where all the
        trailing zeros have been chopped of. Their magnitude is given by the
        length of their representation. If they have the same magnitude the
        lexical order is correct. Bias is used to save that information.
     *)
    let ok = Char.is_digit in
    let bias = ref 0 in
    let rec loop () =
      match next ~ok s1 pos1,next ~ok s2 pos2 with
      | Some _, None -> 1
      | None , Some _ -> -1
      | None , None when !bias <> 0-> !bias
      | None , None ->
          (* Both ints have the same value, The one with the shortest
             representation (i.e. the least trailing zeroes) is
             considered to be the smallest*)
          !pos1 - !pos2
      | Some c1,Some c2 when !bias = 0 ->
          bias := Char.compare c1 c2;
          loop ()
      | Some _ , Some _ -> loop ()
    in
    consume0 s1 pos1;
    consume0 s2 pos2;
    loop ()
  in

  let s1_length = String.length s1 in
  let s2_length = String.length s2 in
  let rec loop () =
    let r  = compare_non_numerical () in
    let r' = compare_numerical () in
    match r,r' with
    | 0,0 when !pos1 = s1_length && !pos2 = s2_length -> 0
    | 0,0 -> loop ()
    | 0,i | i,_ -> i
  in
  loop ()
;;

TEST_MODULE "collate" = struct
  let (<!) s s' = collate s s' < 0

(*
  let (>!) s s' = collate s s' > 0

  let basic_tests = (fun (s,s') ->
    "invertible" @? ((s' <! s) = (s >! s'));
    "total" @? (definitive_clause [s<!s'; s=s'; s>!s']))
*)

  (* repeat 50 basic_tests (pg sg sg);
  repeat 2 basic_tests (dup sg);
  repeat 50 (fun (s,s',s'') ->
    let (s1,s2,s3) =
      match List.sort ~cmp:String.collate [s;s';s''] with
      | [s1;s2;s3] -> s1,s2,s3
      | _ -> assert false
    in
    "transitive" @?
      (((s1 <! s2) || (s2 <! s3)) = (s1 <! s3)))
    (tg sg sg sg); *)

  TEST = "a2b" <! "a10b"
  TEST = "a2b" <! "a02b"
  TEST = "a010b" <! "a20b"

end

(**
   Inverse operation of [String.escaped]
*)
exception Unescape_error of bool*int*string

(* The stdlib's escaped does a lot of fancy wazoo magic to avoid
   using a buffer:
   It works in two passes, the first one calculates the length of the string to
   allocate and the second one does the actual escaping.

   This would be more cumbersome to do here but might be worth the hassle if
   performance ever gets to be an issue *)
let unescaped' ?(strict=true) s =
  let len = String.length s in
  let pos = ref 0 in
  let error ?(fatal=false) message =
    raise (Unescape_error (fatal,!pos,message))
  in
  let consume () =
    let i = !pos in
    if i = len then error "unexpectedly reached end of string";
    let c = s.[i] in
    pos := i + 1;
    c
  in
  let res = Buffer.create len in
  let emit c = Buffer.add_char res c in
  let emit_code code =
    match Char.of_int code with
    | Some c -> emit c
    | None -> error ~fatal:true
        (Printf.sprintf "got invalid escape code %d" code)
  in
  let rec loop () =
    if !pos < len then begin
      let c = consume () in
      if c <> '\\' then
        emit c
      else begin
        let mark = !pos in
        try
          let c = consume () in
          match c with
          | '\\' | '\"' -> emit c
          | 'b' -> emit '\b'
          | 'n' -> emit '\n'
          | 'r' -> emit '\r'
          | 't' -> emit '\t'
          | '\n' ->
              let rec consume_blank () =
                if !pos < len then begin
                  match consume () with
                  | ' ' | '\t' -> consume_blank ()
                  | _ -> decr pos
                end
              in
              consume_blank ()
          | 'x' ->
              let c2hex c =
                if (c >= 'A') && (c <= 'F' ) then
                  (Char.to_int c) + 10 - Char.to_int 'A'
                else if (c >= 'a') && (c <= 'f' ) then
                  (Char.to_int c) + 10 - Char.to_int 'a'
                else if (c >= '0') && (c <= '9') then
                  (Char.to_int c) - Char.to_int '0'
                else
                  error (Printf.sprintf "expected hex digit, got: %c" c);
              in
              let c1 = consume () in
              let c2 = consume () in
              emit_code (16 * c2hex c1 + c2hex c2);
          | c when Char.is_digit c ->
              let char_to_num c =
                match Char.get_digit c with
                | None -> error (Printf.sprintf "expected digit,got: %c" c);
                | Some i -> i
              in
              let i1 = char_to_num c in
              let i2 = char_to_num (consume ()) in
              let i3 = char_to_num (consume ()) in
              emit_code (100 * i1 + 10 * i2 + i3);
          | c -> error (Printf.sprintf "got invalid escape character: %c" c);
        with Unescape_error (false,_,_) when not strict ->
          emit '\\';
          pos := mark
      end;
      loop ()
    end else
      Buffer.contents res;
  in
  loop ();
;;

let unescaped ?strict s =
  try
    unescaped' ?strict s
  with Unescape_error (_,pos,message) ->
    invalid_argf "String.unescaped error at position %d of %s: %s"
      pos s message ()

let unescaped_res ?strict s =
  try
    Core.Result.Ok (unescaped' ?strict s)
  with Unescape_error (_,pos,message) ->
    Core.Result.Error (pos,message)


let squeeze str =
  let len = String.length str in
  let buf = Buffer.create len in
  let rec skip_spaces i =
    if i >= len then
      Buffer.contents buf
    else
      let c = str.[i] in
      if (c = ' ') || (c = '\n') || (c = '\t') || (c = '\r') then
        skip_spaces (i+1)
      else
        begin
          Buffer.add_char buf c;
          copy_chars (i+1)
        end
  and copy_chars i =
    if i >= len then
      Buffer.contents buf
    else
      let c = str.[i] in
      if (c = ' ') || (c = '\n') || (c = '\t') || (c = '\r') then
        begin
          Buffer.add_char buf ' ';
          skip_spaces (i+1)
        end
      else
        begin
          Buffer.add_char buf c;
          copy_chars (i+1)
        end
  in
  copy_chars 0
;;


let pad_right ?(char=' ') s l =
  let src_len = String.length s in
  if src_len >= l then
    s
  else
    let res = String.create l in
    String.blit ~src:s ~dst:res ~src_pos:0 ~dst_pos:0 ~len:src_len;
    String.fill ~pos:src_len ~len:(l-src_len) res char;
    res

let pad_left ?(char=' ') s l =
  let src_len = String.length s in
  if src_len >= l then
    s
  else
    let res = String.create l in
    String.blit ~src:s ~dst:res ~src_pos:0 ~dst_pos:(l-src_len) ~len:src_len;
    String.fill ~pos:0 ~len:(l-src_len) res char;
    res

let line_break ~len s =
  let buf = Buffer.create len in
  let flush_buf () =
    let res = Buffer.contents buf in
    Buffer.reset buf;
    res
  in
  let rec loop acc = function
    | [] ->
        let acc = if Buffer.length buf <> 0 then
          flush_buf ():: acc
        else if acc = [] then
          [""]
        else
          acc
        in
        List.rev acc
    | h::t when Buffer.length buf = 0 ->
        Buffer.add_string buf h;
        loop acc t
    | h::t when (Buffer.length buf + 1 + String.length h) < len ->
        Buffer.add_char buf ' ';
        Buffer.add_string buf h;
        loop acc t
    | l ->
        loop (flush_buf ()::acc) l
  in
  List.concat_map (String.split ~on:'\n' s)
    ~f:(fun s -> loop [] (String.split ~on:' ' s))

(* Finds out where to break a given line; returns the len of the line to break
    and the staring position of the next line.*)
let rec word_wrap__break_one
    ~hard_limit
    ~soft_limit
    ~previous_match
    s
    ~pos
    ~len =
  if pos = String.length s then
    len,pos
  else if previous_match > 0 && len >= soft_limit then
    previous_match,pos-len+previous_match+1
  else if len >= hard_limit then
    len,pos
  else
    match s.[pos] with
      (* Detect \r\n as one newline and not two... *)
    | '\r' when pos < String.length s -1 && s.[pos + 1] = '\n' ->
        len,pos+2
    | '\r' | '\n' -> len,pos+1
    | ' ' | '\t' ->
        word_wrap__break_one s
          ~hard_limit
          ~soft_limit
          ~previous_match:len
          ~pos:(pos+1)
          ~len:(len+1)
    | _ ->
        word_wrap__break_one s
          ~previous_match
          ~hard_limit
          ~soft_limit
          ~pos:(pos+1)
          ~len:(len+1)


(* Returns an pos*length list of all the lines (as substrings of the argument
   passed in) *)
let rec word_wrap__find_substrings ~hard_limit ~soft_limit s acc pos =
  if pos < String.length s then begin
    let len,new_pos = word_wrap__break_one s
      ~hard_limit
      ~soft_limit
      ~previous_match:0
      ~pos
      ~len:0
    in
    word_wrap__find_substrings
      ~hard_limit
      ~soft_limit
      s
      ((pos,len)::acc)
      new_pos
  end else
    acc

let word_wrap
    ?(trailing_nl=false)
    ?(soft_limit=80)
    ?(hard_limit=Int.max_value)
    ?(nl="\n") s =
  let soft_limit = min soft_limit hard_limit in
  let lines = word_wrap__find_substrings ~soft_limit ~hard_limit s [] 0 in
  match lines with
  | [] | [_] ->
      if trailing_nl then s^nl else s
  | ((hpos,hlen)::t) ->
      let nl_len = String.length nl in
      let body_len =
        List.fold_left t
          ~f:(fun acc (_,len) -> acc + nl_len + len)
          ~init:0
      in
      let res_len =
        if trailing_nl then body_len+hlen+nl_len else body_len+hlen
      in
      let res = String.create res_len in
      if trailing_nl then begin
        String.blit
          ~src:nl
          ~dst:res
          ~len:nl_len
          ~src_pos:0
          ~dst_pos:(body_len+hlen);
      end;
      String.blit
        ~src:s
        ~dst:res
        ~len:hlen
        ~src_pos:hpos
        ~dst_pos:body_len;
      let rec blit_loop dst_end_pos = function
        | [] -> ()
        | (src_pos,len)::rest ->
            let dst_pos = dst_end_pos-len-nl_len in
            String.blit
              ~src:s
              ~dst:res
              ~len
              ~src_pos
              ~dst_pos;
            String.blit
              ~src:nl
              ~dst:res
              ~len:nl_len
              ~src_pos:0
              ~dst_pos:(dst_pos + len);
            blit_loop dst_pos rest
      in
      blit_loop body_len t;
      res

(* Knuth-Morris-Pratt string matching. *)
let is_substring_pure_ml  ~substring t =
  let kmp_prefix len ~substring =
    let prefix = Array.create ~len:len 0 in
    let rec f ~k ~q =
      if q > len then prefix
      else (
        let k =
          let rec g k =
            if k <= 0 || ((String.get substring k) = (String.get substring (q - 1))) then k
            else g (prefix.(k - 1))
          in
          g k
        in
        let k =
          if (String.get substring k) = (String.get substring (q - 1))
          then k + 1 else k
        in
        assert (q - 1 >= 0 && q - 1 < len);
        Array.set prefix (q - 1) k;
        f ~k ~q:(q + 1)
      )
    in
    f ~k:0 ~q:2
  in
  let n = String.length t in
  let m = String.length substring in
  let prefix = kmp_prefix m ~substring in
  let rec f ~q ~i =
    if i > n then false
    else
    let q = (
      let q =
        let rec g q =
          if q <= 0 || ((String.get substring q) = (String.get t (i - 1)))
          then q
          else g (prefix.(q - 1))
        in
        g q
      in
      if String.get substring q = String.get t (i - 1) then q + 1
      else q
    )
    in
    if q = m then true
    else f ~q ~i:(i + 1)
  in
  f ~q:0 ~i:1


external is_substring_c_version : string -> string -> bool = "core_extended_is_substring" "noalloc"

let is_substring ~substring:needle haystack =
  (* Hacks for backward compatibility with the ML version that used to be here *)
  if String.length needle = 0 then begin
    if String.length haystack = 0 then false
    else invalid_arg "index out of bounds"
  end else
    is_substring_c_version haystack needle

TEST = is_substring ~substring:"foo" "foo"
TEST = not (is_substring ~substring:"" "")
TEST =
  (* For bug compatibility with the ML version that used to be here *)
  try
    ignore (is_substring ~substring:"" "foo");
    assert false (* should not be reachable *)
  with Invalid_argument _ ->
    true
TEST = not (is_substring ~substring:"foo" "")
TEST = is_substring ~substring:"bar" "foobarbaz"
TEST = not (is_substring ~substring:"Z" "z")
TEST = not (is_substring ~substring:"store" "video stapler")
TEST = not (is_substring ~substring:"sandwich" "apple")
TEST = is_substring ~substring:"z" "abc\x00z"

BENCH "is_substring C version" =
  assert (is_substring ~substring:"xyz" "abcdefghijklmnopqrstuvwxyz");
BENCH "is_substring ML version" =
  assert (is_substring_pure_ml ~substring:"xyz" "abcdefghijklmnopqrstuvwxyz")

let edit_distance_matrix ?transpose s1 s2 =
  let transpose = Option.is_some transpose in
  let l1, l2 = String.length s1, String.length s2 in
  let d = Array.make_matrix 0 ~dimx:(l1+1) ~dimy:(l2+1) in
  for x=0 to l1 do d.(x).(0) <- x done;
  for y=0 to l2 do d.(0).(y) <- y done;
  for y=1 to l2 do
    for x=1 to l1 do
      let min_d =
        if s1.[x-1] = s2.[y-1] then d.(x-1).(y-1)
        else List.reduce_exn ~f:min
          [d.(x-1).(y) + 1;
           d.(x).(y-1) + 1;
           d.(x-1).(y-1) + 1]
      in
      let min_d =
        if transpose && x > 1 && y > 1
          && s1.[x-1] = s2.[y-2] && s1.[x-2] = s2.[y-1]
        then min min_d (d.(x-2).(y-2) + 1)
        else min_d
      in
      d.(x).(y) <- min_d
    done;
  done;
  d

let edit_distance ?transpose s1 s2 =
  (edit_distance_matrix ?transpose s1 s2).(String.length s1).(String.length s2)

TEST = edit_distance "" "" = 0
TEST = edit_distance "stringStringString" "stringStringString" = 0
TEST = edit_distance "ocaml" "coaml" = 2
TEST = edit_distance ~transpose:() "ocaml" "coaml" = 1
TEST = edit_distance "sitting" "kitten" = 3
TEST = edit_distance ~transpose:() "sitting" "kitten" = 3
TEST = edit_distance "abcdef" "1234567890" = 10
TEST = edit_distance "foobar" "fubahr" = 3
TEST = edit_distance "hylomorphism" "zylomorphism" = 1
